from __future__ import annotations

from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from aasm.intermediate.argument import Argument


class Instruction:
    def __init__(self, **kwargs: Dict[str, Argument]):
        for key in kwargs:
            self.__dict__[key] = kwargs[key]
  
    def print(self) -> None:
        print('Instruction')
        for argument in self.__dict__.values():
            argument.print()


class IfGreaterThan(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)
 
    def print(self) -> None:
        print('IfGreaterThan')
        super().print()


class IfGreaterThanOrEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)
 
    def print(self) -> None:
        print('IfGreaterThanOrEqual')
        super().print()


class IfLessThan(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('IfLessThan')
        super().print()


class IfLessThanOrEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('IfLessThanOrEqual')
        super().print()


class IfEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('IfEqual')
        super().print()


class IfNotEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('IfNotEqual')
        super().print()


class WhileGreaterThan(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('WhileGreaterThan')
        super().print()


class WhileGreaterThanOrEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('WhileGreaterThanOrEqual')
        super().print()


class WhileLessThan(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('WhileLessThan')
        super().print()


class WhileLessThanOrEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('WhileLessThanOrEqual')
        super().print()


class WhileEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('WhileEqual')
        super().print()


class WhileNotEqual(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('WhileNotEqual')
        super().print()


class Multiply(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('Multiply')
        super().print()


class Divide(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('Divide')
        super().print()


class Add(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('Add')
        super().print()


class Subtract(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('Subtract')
        super().print()


class AddElement(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('AddElement')
        super().print()


class RemoveElement(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('RemoveElement')
        super().print()


class Subset(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument, arg3: Argument):
        super().__init__(arg1=arg1, arg2=arg2, arg3=arg3)

    def print(self) -> None:
        print('RemoveElement')
        super().print()


class Set(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('Set')
        super().print()


class IfInList(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('IfInList')
        super().print()


class IfNotInList(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('IfNotInList')
        super().print()


class Clear(Instruction):
    def __init__(self, arg1: Argument):
        super().__init__(arg1=arg1)

    def print(self) -> None:
        print('Clear')
        super().print()


class Length(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('Length')
        super().print()


class Send(Instruction):
    def __init__(self, arg1: Argument):
        super().__init__(arg1=arg1)

    def print(self) -> None:
        print('Send')
        super().print()
        
        
class RemoveNElements(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('RemoveNElements')
        super().print()


class UniformDist(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument, arg3: Argument):
        super().__init__(arg1=arg1, arg2=arg2, arg3=arg3)

    def print(self) -> None:
        print('UniformDist')
        super().print()


class NormalDist(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument, arg3: Argument):
        super().__init__(arg1=arg1, arg2=arg2, arg3=arg3)

    def print(self) -> None:
        print('NormalDist')
        super().print()


class ExpDist(Instruction):
    def __init__(self, arg1: Argument, arg2: Argument):
        super().__init__(arg1=arg1, arg2=arg2)

    def print(self) -> None:
        print('ExpDist')
        super().print()


class Round(Instruction):
    def __init__(self, arg1: Argument):
        super().__init__(arg1=arg1)

    def print(self) -> None:
        print('Round')
        super().print()



from __future__ import annotations

from typing import List


class PythonCode:
    def __init__(self, indent_size: int):
        self.indent_size = indent_size
        self.indent: int = 0
        self.code_lines: List[str] = []

    def indent_left(self) -> None:
        self.indent -= self.indent_size
        
    def indent_right(self) -> None:
        self.indent += self.indent_size
        
    def add_line(self, line: str) -> None:
        self.code_lines.append(self.indent * ' ' + line + '\n')
        
    def add_newline(self) -> None:
        self.add_line('')
        
    def add_newlines(self, count: int) -> None:
        for _ in range(count):
            self.add_newline()



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.generating.python_code import PythonCode
from aasm.intermediate.graph import (AgentConstantAmount, AgentPercentAmount,
                                     ConnectionConstantAmount,
                                     ConnectionDistNormalAmount,
                                     StatisticalGraph)

if TYPE_CHECKING:
    from aasm.intermediate.graph import Graph


class PythonGraph(PythonCode):
    def __init__(self, indent_size: int, graph: Graph):
        super().__init__(indent_size)
        if graph:
            self.add_required_imports()
            self.add_newlines(2)
            self.generate_graph(graph)

    def add_required_imports(self) -> List[str]:
        self.add_line('import random')
        self.add_line('import uuid')
        self.add_line('import numpy')
        
    def generate_graph(self, graph: Graph) -> None:
        if isinstance(graph, StatisticalGraph):
            self.add_statistical_graph(graph)

    def add_statistical_graph(self, graph: StatisticalGraph) -> None:
        self.add_line('def generate_graph_structure(domain):')
        self.indent_right()

        if not graph.agents:
            self.add_line('return []')
            self.indent_left()
            return
        
        num_agents = []
        for agent in graph.agents.values():
            if isinstance(agent.amount, AgentConstantAmount):
                self.add_line(f'_num_{agent.name} = {agent.amount.value}')
            elif isinstance(agent.amount, AgentPercentAmount):
                self.add_line(f'_num_{agent.name} = round({agent.amount.value} / 100 * {graph.size})')
            num_agents.append(f'_num_{agent.name}')
        self.add_line(f'num_agents = {" + ".join(num_agents)}')

        self.add_line('random_id = str(uuid.uuid4())[:4]')
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]')

        self.add_line('agents = []')
        self.add_line('next_agent_idx = 0')
        for agent in graph.agents.values():
            self.add_line(f'for _ in range(_num_{agent.name}):')
            self.indent_right()
            if isinstance(agent.connections, ConnectionDistNormalAmount):
                self.add_line(f'num_connections = int(numpy.random.normal({agent.connections.mean}, {agent.connections.std_dev}))')
            elif isinstance(agent.connections, ConnectionConstantAmount):
                self.add_line(f'num_connections = {agent.connections.value}')
            self.add_line('num_connections = max(min(num_connections, len(jids)), 0)')
            self.add_line('agents.append({')
            self.indent_right()
            self.add_line('"jid": jids[next_agent_idx],')
            self.add_line(f'"type": "{agent.name}",')
            self.add_line('"connections": random.sample(jids, num_connections),')
            self.indent_left()
            self.add_line('})')
            self.add_line('next_agent_idx += 1')
            self.indent_left()
        
        self.add_line('return agents')
        self.indent_left()



import copy
import datetime
import json
import random
import httpx
import numpy
import spade


class average_user(spade.agent.Agent):
    def __init__(self, jid, password, connections, backup_url = None, backup_period = 60, backup_delay = 0, logger = None, **kwargs):
        super().__init__(jid, password, verify_security=False)
        if logger: logger.debug(f"[{jid}] Received parameters: jid: {jid}, password: {password}, connections: {connections}, backup_url: {backup_url}, backup_period: {backup_period}, backup_delay: {backup_delay}, kwargs: {kwargs}")
        self.logger = logger
        self.connections = connections
        self.backup_url = backup_url
        self.backup_period = backup_period
        self.backup_delay = backup_delay
        self.msgRCount = kwargs.get("msgRCount", 0)
        self.msgSCount = kwargs.get("msgSCount", 0)
        self.friends = kwargs.get("friends", [])
        if self.logger: self.logger.debug(f"[{self.jid}] Class dict after initialization: {self.__dict__}")
    
    @property
    def connCount(self):
        return len(self.connections)
    
    def get_json_from_spade_message(self, msg):
        return json.loads(msg.body)
    
    def get_spade_message(self, receiver_jid, body):
        msg = spade.message.Message(to=receiver_jid)
        body["sender"] = str(self.jid)
        msg.metadata["type"] = body["type"]
        msg.metadata["performative"] = body["performative"]
        msg.body = json.dumps(body)
        return msg
    
    def setup(self):
        if self.backup_url:
            self.add_behaviour(self.BackupBehaviour(start_at=datetime.datetime.now() + datetime.timedelta(seconds=self.backup_delay), period=self.backup_period))
        self.add_behaviour(self.facebook_activity(period=30))
        if self.logger: self.logger.debug(f"[{self.jid}] Class dict after setup: {self.__dict__}")
    
    class BackupBehaviour(spade.behaviour.PeriodicBehaviour):
        async def run(self):
            data = {
                "jid": str(self.agent.jid),
                "connections": self.agent.connections,
                "connCount": self.agent.connCount,
                "msgRCount": self.agent.msgRCount,
                "msgSCount": self.agent.msgSCount,
                "friends": self.agent.friends,
            }
            if self.agent.logger: self.agent.logger.debug(f"[{self.agent.jid}] Sending backup data: {data}")
            try:
                async with httpx.AsyncClient() as client:
                    await client.post(self.agent.backup_url, json=data)
            except Exception as e:
                if self.agent.logger: self.agent.logger.warn(f"[{self.agent.jid}] Backup error: {e}")
    
    class facebook_activity(spade.behaviour.PeriodicBehaviour):
        async def post_photos(self):
            if self.agent.logger: self.agent.logger.debug(f"[{self.agent.jid}] Run action post_photos")
            send = { "type": "facebook_post", "performative": "query", "photos": 0.0, }
            num_photos = 0
            num_photos = numpy.random.normal(21, 37)
            num_photos = round(num_photos)
            send["photos"] = num_photos
            if self.agent.logger: self.agent.logger.debug(f"[{self.agent.jid}] Send message {send} to self.agent.friends")
            for receiver in self.agent.friends:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount += 1
        
        async def run(self):
            await self.post_photos()
    

import random
import uuid
import numpy


def generate_graph_structure(domain):
    _num_average_user = round(100 / 100 * 150)
    num_agents = _num_average_user
    random_id = str(uuid.uuid4())[:4]
    jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]
    agents = []
    next_agent_idx = 0
    for _ in range(_num_average_user):
        num_connections = int(numpy.random.normal(0, 15))
        num_connections = max(min(num_connections, len(jids)), 0)
        agents.append({
            "jid": jids[next_agent_idx],
            "type": "average_user",
            "connections": random.sample(jids, num_connections),
        })
        next_agent_idx += 1
    return agents



from __future__ import annotations

from typing import Dict


class AgentAmount:
    def __init__(self, value: str):
        self.value = value

    def print(self) -> None:
        print(f'AgentAmount value = {self.value}')


class AgentConstantAmount(AgentAmount):
    def __init__(self, value: str):
        super().__init__(value)

    def print(self) -> None:
        print('AgentConstantAmount')
        super().print()


class AgentPercentAmount(AgentAmount):
    def __init__(self, value: str):
        super().__init__(value)

    def print(self) -> None:
        print('AgentPercentAmount')
        super().print()


class ConnectionAmount:
    ...


class ConnectionConstantAmount(ConnectionAmount):
    def __init__(self, value: str):
        self.value = value
    
    def print(self) -> None:
        print(f'ConnectionConstantAmount value = {self.value}')


class ConnectionDistNormalAmount(ConnectionAmount):
    def __init__(self, mean: str, std_dev: str):
        self.mean = mean
        self.std_dev = std_dev

    def print(self) -> None:
        print(f'ConnectionDistNormalAmount mean = {self.mean}, std_dev = {self.std_dev}')


class StatisticalAgent:
    def __init__(self, name: str, amount: AgentAmount, connections: ConnectionAmount):
        self.name = name
        self.amount = amount
        self.connections = connections

    def print(self) -> None:
        print(f'StatisticalAgent name = {self.name}')
        self.amount.print()
        self.connections.print()


class Graph:
    def __init__(self):
        self.size = None

    def set_size(self, size: int) -> None:
        self.size = size

    def is_size_defined(self) -> bool:
        return self.size is not None

    def print(self) -> None:
        print(f'Graph size = {self.size}')


class StatisticalGraph(Graph):
    def __init__(self):
        super().__init__()
        self.agents: Dict[str, StatisticalAgent] = {}

    def add_agent(self, agent: StatisticalAgent) -> None:
        self.agents[agent.name] = agent

    def is_agent_defined(self, agent_type: str) -> bool:
        return agent_type in self.agents

    def is_agent_percent_amount_used(self) -> bool:
        for agent in self.agents.values():
            if isinstance(agent.amount, AgentPercentAmount):
                return True
        return False

    def print(self) -> None:
        super().print()
        print('StatisticalGraph')
        for agent in self.agents.values():
            agent.print()



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.parsing.op.action import op_ACTION, op_EACTION
from aasm.parsing.op.agent import op_AGENT, op_EAGENT
from aasm.parsing.op.behav import op_BEHAV, op_EBEHAV
from aasm.parsing.op.clr import op_CLR
from aasm.parsing.op.conditional import (
    handle_ordered_conditional_statement,
    handle_unordered_conditional_statement)
from aasm.parsing.op.decl import op_DECL
from aasm.parsing.op.defg import op_DEFG
from aasm.parsing.op.eblock import op_EBLOCK
from aasm.parsing.op.graph import op_EGRAPH, op_GRAPH
from aasm.parsing.op.len import op_LEN
from aasm.parsing.op.list import (handle_list_inclusion,
                                  handle_list_modification)
from aasm.parsing.op.math import handle_math_statement
from aasm.parsing.op.message import op_EMESSAGE, op_MESSAGE
from aasm.parsing.op.prm import op_agent_PRM, op_message_PRM
from aasm.parsing.op.rand import op_RAND
from aasm.parsing.op.remen import op_REMEN
from aasm.parsing.op.round import op_ROUND
from aasm.parsing.op.send import op_SEND
from aasm.parsing.op.set import op_SET
from aasm.parsing.op.size import op_SIZE
from aasm.parsing.op.subs import op_SUBS
from aasm.parsing.state import State

if TYPE_CHECKING:
    from aasm.parsing.state import ParsedData


def parse_lines(lines: List[str], debug: bool) -> ParsedData:
    state = State(lines, debug)
    for tokens in state.tokens_from_lines():
        match tokens:                                        
            case [ 'AGENT', name ]:
                op_AGENT(state, name)
            
            case [ 'EAGENT' ]:
                op_EAGENT(state)
                
            case [ 'MESSAGE', name, performative ]:
                op_MESSAGE(state, name, performative)
            
            case [ 'EMESSAGE' ]:
                op_EMESSAGE(state)
                
            case [ 'PRM', name, category ]:
                op_message_PRM(state, name, category)
                
            case [ 'PRM', name, category, *args ]:
                op_agent_PRM(state, name, category, args)
                
            case [ 'BEHAV', name, category, *args ]:
                op_BEHAV(state, name, category, args)
                
            case [ 'EBEHAV' ]:
                op_EBEHAV(state)
                
            case [ 'ACTION', name, category, *args ]:
                op_ACTION(state, name, category, args)
                
            case [ 'EACTION' ]:
                op_EACTION(state)
                
            case [ 'DECL', name, value ]:
                op_DECL(state, name, value)
                
            case [ 'EBLOCK' ]:
                op_EBLOCK(state)
                
            case [ 'IEQ' | 'INEQ' | 'WEQ' | 'WNEQ' as op, arg1, arg2 ]:
                handle_unordered_conditional_statement(state, op, arg1, arg2)
                
            case [ 'IGT' | 'IGTEQ' | 'ILT' | 'ILTEQ' | 'WGT' | 'WGTEQ' | 'WLT' | 'WLTEQ' as op, arg1, arg2 ]:
                handle_ordered_conditional_statement(state, op, arg1, arg2)
                
            case [ 'ADD' | 'SUBT' | 'MULT' | 'DIV' as op, arg1, arg2 ]:
                handle_math_statement(state, op, arg1, arg2)
                
            case [ 'ADDE' | 'REME' as op, list_, element ]:
                handle_list_modification(state, op, list_, element)
                
            case [ 'LEN', result, list_ ]:
                op_LEN(state, result, list_)
                
            case [ 'CLR', list_ ]:
                op_CLR(state, list_)
                
            case [ 'IN' | 'NIN' as op, list_, element ]:
                handle_list_inclusion(state, op, list_, element)
                
            case [ 'SEND', rcv_list ]:
                op_SEND(state, rcv_list)
                
            case [ 'SUBS', dst_list, src_list, num ]:
                op_SUBS(state, dst_list, src_list, num)
                
            case [ 'SET', arg1, arg2 ]:
                op_SET(state, arg1, arg2)
                
            case [ 'REMEN', list_, num ]:
                op_REMEN(state, list_, num)

            case [ 'RAND', result, cast_to, dist, *args ]:
                op_RAND(state, result, cast_to, dist, args)

            case [ 'ROUND', num ]:
                op_ROUND(state, num)
                
            case [ 'SET', arg1, arg2 ]:
                op_SET(state, arg1, arg2)

            case [ 'GRAPH', category ]:
                op_GRAPH(state, category)

            case [ 'EGRAPH' ]:
                op_EGRAPH(state)

            case [ 'SIZE', size ]:
                op_SIZE(state, size)

            case [ 'DEFG',  agent_name, amount, *args ]:
                op_DEFG(state, agent_name, amount, args)
                
            case _:
                state.panic(f'Unknown tokens: {tokens}')

    return state.get_parsed_data()



from __future__ import annotations

from typing import List


class Code:
    def __init__(self, agent_code_lines: List[str], graph_code_lines: List[str]):
        self.agent_code_lines: List[str] = agent_code_lines
        self.graph_code_lines: List[str] = graph_code_lines
        self._iter_code_lines: List[str] | None = None
        self._iter_idx: int | None = None
        self._iter_num_code_lines: int | None = None

    def __iter__(self) -> Code:
        self._iter_code_lines = self.agent_code_lines + ["\n"] + self.graph_code_lines
        self._iter_num_code_lines = len(self._iter_code_lines)
        self._iter_idx = -1
        return self

    def __next__(self):
        self._iter_idx += 1
        if self._iter_idx < self._iter_num_code_lines:
            return self._iter_code_lines[self._iter_idx]
        raise StopIteration



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.action import SendMessageAction
from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Send

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_SEND(state: State, arg1: str) -> None:
    state.require(state.in_action, 'Not inside any action.', 'SEND can be used inside send_msg actions.')
    state.require(
        isinstance(state.last_action, SendMessageAction), 
        'Not inside send_msg action.', 
        'SEND can be used inside send_msg actions.'
    )
    receivers = Argument(state, arg1)
    state.require(receivers.send_context(), 'Mismatched type in the send context.', f'ARG1 {receivers.explain()}')

    state.last_action.add_instruction(Send(receivers))



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Clear

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_CLR(state: State, arg1: str) -> None:
    state.require(state.in_action, 'Not inside any action.', 'CLR can be used inside actions.')
    list_ = Argument(state, arg1)
    state.require(list_.list_clear_context(), 'Mismatched type in the list clear context.', f'ARG {list_.explain()}')

    state.last_action.add_instruction(Clear(list_))
